package com.brave.servlet;

import com.brave.annotation.MyController;
import com.brave.annotation.MyRequestMapping;
import org.reflections.Reflections;

import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.*;

public class MyDispatcherServlet extends HttpServlet {
    /**
     * 配置扫描的包，放到一个.properties文件中，在初始化的时候读取
     */
    private Properties properties=new Properties();

    /**
     * 需要一个set，把所有能够相应controller的存起来
     */
    private Set<Class<?>> classSet=new HashSet<>();

    /**
     * SpringMVC容器
     */
    private Map<String, Object> springMVCContext=new HashMap<>();


    /**
     * 映射处理器，存储所有的方法
     */
    private Map<String, Method> handlerMapping=new HashMap<>();

    /**
     * 后端处理器的映射关系，存储所有controller
     */
    private Map<String,Object> controllerMap=new HashMap<>();


    @Override
    public void init(ServletConfig config) throws ServletException {
        //加载配置文件，在web.xml中配置的初始化参数contextfigLocation
        String initParameter = config.getInitParameter("contextConfigLocation");
        try {
            loadConfigFile(initParameter);
        }catch (Exception e){
            e.printStackTrace();
        }

        //扫描controller
        scanPackage(properties.getProperty("package"));

        //初始化controller包
        initController();
        //初始化处理器映射器
        initHandlerMapping();
    }

    @Override
    protected void service(HttpServletRequest req, HttpServletResponse resp){
        //处理请求
        if (handlerMapping.isEmpty()){
            return;
        }

        //获取url
        String uri = req.getRequestURI();
        String contextPath = req.getContextPath();
        String url = uri.replace(contextPath, "");
        if (!handlerMapping.containsKey(url)){
            try {
                resp.getWriter().println("404");
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        }else {
            Method method = handlerMapping.get(url);
            Object controller = controllerMap.get(url);
            try {
                method.invoke(controller);
            } catch (IllegalAccessException e) {
                throw new RuntimeException(e);
            } catch (InvocationTargetException e) {
                throw new RuntimeException(e);
            }
        }
    }

    private void initHandlerMapping() {
        if (springMVCContext.isEmpty()){
            return;
        }

        for (Map.Entry<String, Object> entry : springMVCContext.entrySet()) {
            //获取class对象
            Class<?> aClass = entry.getValue().getClass();
            if (!aClass.isAnnotationPresent(MyController.class)){
                continue;
            }else {
                String baseUrl="";
                if (aClass.isAnnotationPresent(MyRequestMapping.class)){
                    //如果类包含注解MyRequestMapping，获取注解值
                    MyRequestMapping annotation = aClass.getAnnotation(MyRequestMapping.class);
                    baseUrl=annotation.value();
                }

                //获取所有方法
                Method[] methods=aClass.getMethods();
                for (Method method : methods) {
                    //判断方法上含有MyRequestMapping注解
                    if (method.isAnnotationPresent(MyRequestMapping.class)){
                        MyRequestMapping annotation = method.getAnnotation(MyRequestMapping.class);
                        String url = annotation.value();
                        url = baseUrl +url;

                        //将该放入方法集
                        handlerMapping.put(url,method);

                        try {
                            //放入controllerMap中
                            controllerMap.put(url,aClass.newInstance());
                        } catch (InstantiationException e) {
                            throw new RuntimeException(e);
                        } catch (IllegalAccessException e) {
                            throw new RuntimeException(e);
                        }
                    }
                }
            }
        }
    }

    private void initController() {
        if (classSet.isEmpty()){
            return;
        }
        for (Class<?> controller : classSet) {
            try {
                springMVCContext.put(lowerFirstWord(controller.getSimpleName()), controller.newInstance());
            } catch (InstantiationException e) {
                throw new RuntimeException(e);
            } catch (IllegalAccessException e) {
                throw new RuntimeException(e);
            }
        }
    }

    private String lowerFirstWord(String simpleName) {
        char[] array = simpleName.toCharArray();
        array[0] +=32;
        return String.valueOf(array);
    }

    private void scanPackage(String packageName) {
        Reflections reflections = new Reflections(packageName);
        classSet=reflections.getTypesAnnotatedWith(MyController.class);
    }


    private void loadConfigFile(String fileName) {
        //以流的方式获取资源
        InputStream inputStream = this.getClass().getClassLoader().getResourceAsStream(fileName);

        try {
            properties.load(inputStream);
        } catch (IOException e) {
            throw new RuntimeException(e);
        }finally {
            try {
                inputStream.close();
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        }

    }
}
